/* 	Copyright 2000 Be Incorporated. All Rights Reserved.
**	This file may be used under the terms of the Be Sample Code 
**	License.
*/

#include <OS.h>
#include <Drivers.h>
#include <KernelExport.h>
#include <stdlib.h>
#include <stdio.h>

#include <USB.h>
#include <USB_spec.h>

#define DEBUG_DRIVER 1

#if DEBUG_DRIVER
#define DPRINTF(x) dprintf x
#else
#define DPRINTF(x) ((void)0)
#endif

int32 api_version = B_CUR_DRIVER_API_VERSION;

///* ~44 kHz 16bit stereo */
//#define SAMPLE_SIZE	4  
//#define QS_OUT	(44*SAMPLE_SIZE*TS)
//
//#define TS	4	/* 4 ms buffers */
//#define NB	2
//
//int qs_out = QS_OUT;
//
typedef struct scannerdev scannerdev;
//typedef struct iso_channel iso_channel;	
//typedef struct iso_packet iso_packet;
//
//struct iso_packet
//{
//	struct iso_packet*		next;
//	struct iso_channel*		channel;	
//	void*					buffer;
//	uint32					status;
//	size_t					buffer_size;
//	rlea*					rle_array;
//};
//
//struct iso_channel
//{
//	usb_pipe*	ep;
//	
//	iso_packet*	next_to_queue;
//	iso_packet*	current_rw;
//	size_t		remain;
//	char*		buf_ptr;
//	sem_id		num_available_packets;
//	area_id		buffer_area;
//	iso_packet	iso_packets[NB];
//	
//	int active;
//};

struct scannerdev 
{
	scannerdev *next;
	
	int open;
	int number;
	int fake_fd;
	const usb_device *dev;
	usb_pipe *in, *out;
	// used to implement sync I/O
	sem_id lock;
	sem_id notify;
	area_id buffer_id;
	void *buffer;

	size_t actual;
	status_t status;
	
//	iso_channel	out_channel;
};

/* handy strings for referring to ourself */
#define ID "usb_scanner: "
static const char *drivername = "usb_scanner";
static const char *basename = "scanner/usb";

/* list of device instances and names for publishing */
static scannerdev *device_list = NULL;
static sem_id dev_list_lock = -1;
static int device_count = 0;

static char **device_names = NULL;

/* handles for the USB bus manager */
static char *usb_name = B_USB_MODULE_NAME;
static usb_module_info *usb;


/* Device addition and removal ---------------------------------------
**
** add_device() and remove_device() are used to create and tear down
** device instances.  They are driver by the callbacks device_added()
** and device_removed() which are invoked by the USB bus manager.
*/

static scannerdev * 
add_device(const usb_device *dev, const usb_configuration_info *conf)
{
	scannerdev *ad = NULL;
	int num,i,ifc,alt;
	const usb_interface_info *ii;
	status_t st;
	usb_pipe *in,*out;
	
	DPRINTF((ID "add_device(%p, %p)\n", dev, conf));

	num=0;	
//	if((num = get_number()) < 0) {
//		return NULL;
//	}
	
	for(ifc = 0; ifc < conf->interface_count; ifc++)
	{
		for(alt = 0; alt < conf->interface[ifc].alt_count; alt++) 
		{
			ii = &conf->interface[ifc].alt[alt];

//			/* does it have an AudioStreaming interface? */
//			if(ii->descr->interface_class != AC_AUDIO) continue;
//			if(ii->descr->interface_subclass != AC_AUDIOSTREAMING) continue;
//			if(ii->endpoint_count != 1) continue;
//			if(ii->endpoint_count!=2 || ii->endpoint_count!=3) continue;			
			/* ignore input endpoints */
//			if(ii->endpoint[0].descr->endpoint_address & 0x80) continue;
			
			/* does it support 2 channel, 16 bit audio? */
//			for(i = 0; i < ii->generic_count; i++){
//				usb_format_type_descr *ft = (usb_format_type_descr*) ii->generic[i];
//				if(ft->type != AC_CS_INTERFACE) continue;
//				if(ft->subtype != AC_FORMAT_TYPE) continue;
//				if(ft->length < sizeof(usb_format_type_descr)) continue;
//				if(ft->num_channels != 2) continue;
//				if(ft->subframe_size != 2) continue;
//				if(ft->sample_freq_type != 0) continue;
//				
//				dprintf(ID "found a 2ch, 16bit isoch output (ifc=%d, alt=%d, mp=%d)\n",
//						ifc, alt, ii->endpoint[0].descr->max_packet_size);
				
				goto got_one;
//			}
		}
	}
	
fail:
//	put_number(num);
	if(ad) free(ad);
	return NULL;

got_one:			
	if((ad = (scannerdev *) malloc(sizeof(scannerdev))) == NULL) goto fail;
	
	ad->dev = dev;
	ad->number = num;
	ad->open = 0;
	ad->fake_fd=-2;
	
	if((st = usb->set_alt_interface(dev, ii)) != B_OK) {
		dprintf(ID "set_alt_interface(0) returns %d\n", st);
		goto fail;
	}

	if((st = usb->set_configuration(dev,conf)) != B_OK) {
		dprintf(ID "set_configuration() returns %d\n", st);
		goto fail;
	}

	for(i=0; i<ii->endpoint_count;i++)
	{
		dprintf(ID "endpoint %ld, attr %ld, addr %ld", i,ii->endpoint[i].descr->attributes,ii->endpoint[i].descr->endpoint_address);
		if((ii->endpoint[i].descr->attributes==2) && (ii->endpoint[i].descr->endpoint_address & 0x80)) in = ii->endpoint[i].handle;
		if((ii->endpoint[i].descr->attributes==2) && ((ii->endpoint[i].descr->endpoint_address & 0x80)==0)) out = ii->endpoint[i].handle;		
	}
	ad->in=in;
	ad->out=out;	

	if((ad->lock = create_sem(1,"usb rawdev lock")) < 0) goto fail;
	
	if((ad->notify = create_sem(0,"usb rawdev notify")) < 0) goto fail;
	
	if((ad->buffer_id = create_area("usb rawdev area", &(ad->buffer),
								   B_ANY_KERNEL_ADDRESS, 65536, B_CONTIGUOUS,
								   B_READ_AREA | B_WRITE_AREA)) < 0) goto fail;
								   
	DPRINTF((ID "added %p (out=%p) (/dev/%s%d)\n", ad, out, basename, num));
	
//	if((st = usb->set_pipe_policy(out, NB, TS, SAMPLE_SIZE)) != B_OK){
//		dprintf(ID "set_pipe_policy(out) returns %d\n", st);
//		goto fail;
//	}
//
//	init_iso_channel(&ad->out_channel, out, qs_out, FALSE);
//	start_iso_channel_out(&ad->out_channel);

	/* add it to the list of devices so it will be published, etc */	
	acquire_sem(dev_list_lock);
	ad->next = device_list;
	device_list = ad;
	device_count++;
	release_sem(dev_list_lock);
	
	return ad;
}

static void 
remove_device(scannerdev *ad)
{
//	if(ad->out_channel.active) stop_iso_channel(&ad->out_channel);
//	uninit_iso_channel(&ad->out_channel);
//	put_number(ad->number);

//	usb->cancel_queued_transfers(ad->in); 
//	usb->cancel_queued_transfers(ad->out); 

	delete_sem(ad->lock);
	delete_sem(ad->notify);
	delete_area(ad->buffer_id);

	free(ad);
}

static status_t 
device_added(const usb_device *dev, void **cookie)
{
	const usb_configuration_info *conf;
	scannerdev *ad;
	int i;

	DPRINTF((ID "device_added(%p,...)\n", dev));

	if(conf = usb->get_nth_configuration(dev, 0)) 
	{
		if((ad = add_device(dev, conf)) != NULL)
		{
			*cookie = (void*) ad;
			return B_OK;
		}
	}
	return B_ERROR;
}


static status_t 
device_removed(void *cookie)
{
	scannerdev *ad = (scannerdev *) cookie;
	int i;

	DPRINTF((ID "device_removed(%p)\n",ad));
	
	acquire_sem(dev_list_lock);

	/* mark it as inactive and encourage IO to finish */	
//	ad->out_channel.active = 0;
//	delete_sem(ad->out_channel.num_available_packets);

	/* remove it from the list of devices */	
	if(ad == device_list){
		device_list = ad->next;
	} else {
		scannerdev *n;
		for(n = device_list; n; n = n->next){
			if(n->next == ad){
				n->next = ad->next;
				break;
			}
		}
	}
	device_count--;
	
	/* tear it down if it's not open -- 
	   otherwise the last device_free() will handle it */
		
	// remove the fake fd, to reach a real 0 count
	if(ad->fake_fd>=0) 
	{
		int fd=ad->fake_fd;
		ad->fake_fd=-1;
		close(fd);
	}
	
	if(ad->open == 0){
		remove_device(ad);
	} else {
		DPRINTF((ID "device /dev/%s%d still open -- marked for removal\n",
				 basename,ad->number));
	}
	
	release_sem(dev_list_lock);
	
	return B_OK;
}

/* Device Hooks -----------------------------------------------------------
**
** Here we implement the posixy driver hooks (open/close/read/write/ioctl)
*/

static status_t
device_open(const char *dname, uint32 flags, void **cookie)
{	
	scannerdev *ad=device_list;
	int n;
	
	n = atoi(dname + strlen(basename)); 

	DPRINTF((ID "device_open(\"%s\",%d,...)\n",dname,flags));
	
	acquire_sem(dev_list_lock);
	for(ad = device_list; ad; ad = ad->next)
	{
//		if(ad->number == n)
		{
//			if(ad->out_channel.active) 
			{
				ad->open++;
				*cookie = ad;	
				release_sem(dev_list_lock);
				return B_OK;
			} 
//			else 
//			{
//				dprintf(ID "device is going offline. cannot open\n");
//			}
		}
	}
	release_sem(dev_list_lock);
	return B_ERROR;
}
 
static status_t
device_close (void *cookie)
{
	scannerdev *ad = (scannerdev*) cookie;
	DPRINTF((ID "device_close() name = \"%s%d\"\n",basename,ad->number));
	return B_OK;
}

static status_t
device_free(void *cookie)
{
	scannerdev *ad = (scannerdev *) cookie;
	
	DPRINTF((ID "device_free() name = \"%s%d\"\n",basename,ad->number));
	
	acquire_sem(dev_list_lock);
	ad->open--;
	if((ad->open == 0) && (ad->fake_fd == -1)) 
	{
		remove_device(ad);
	}
	release_sem(dev_list_lock);
	
	return B_OK;
}

// I/O is async, this one releases the lock after the op has completed

void cb_control(void *cookie, uint32 status, void *data, uint32 actual_len)
{
	scannerdev *d = (scannerdev *) cookie;
	d->actual = actual_len;
	d->status = (status>0) ? -1 : 0;
	release_sem(d->notify);
}
				
static status_t
device_read(void *cookie, off_t pos, void *buf, size_t *count)
{
	scannerdev* ad = (scannerdev*) cookie; 
	status_t status;
	
	if(*count >65536) return(B_ERROR);
	
	acquire_sem(ad->lock);
//	memcpy(ad->buffer,buf,*count);
	
	DPRINTF((ID "device_read(%p,%Ld,0x%x,%d) name = \"%s%d\"\n",
			 cookie, pos, buf, *count, basename, ad->number));
//	snooze(50000);
	status=usb->queue_bulk(ad->in,ad->buffer,*count,cb_control,cookie);

	if(!status)
	{
		if(acquire_sem_etc(ad->notify,1,B_CAN_INTERRUPT|B_RELATIVE_TIMEOUT,100000000))
		{
			/* timeout or abort */
			usb->cancel_queued_transfers(ad->in);
			DPRINTF((ID "INTERRUPTED OR TIMEOUT\n"));
			acquire_sem(ad->notify);
			ad->actual=0;
		}
		DPRINTF((ID " data read(%ld)\n",ad->actual ));
		memcpy(buf,ad->buffer,ad->actual);
		*count = ad->actual;
	}
		release_sem(ad->lock);

	return ((ad->status>0) ? -1 : 0);
}

static status_t
device_write(void *cookie, off_t pos, const void *buf, size_t *count)
{
	scannerdev* ad = (scannerdev*) cookie; 
//	iso_channel* channel = &ad->out_channel;
	status_t status;

	if(*count >65536) return(B_ERROR);
	
	acquire_sem(ad->lock);
	memcpy(ad->buffer,buf,*count);
	
	DPRINTF((ID "device_write(%p,%Ld,0x%x,%d) name = \"%s%d\"\n",
			 cookie, pos, buf, *count, basename, ad->number));
//	snooze(50000);
	status=usb->queue_bulk(ad->out,ad->buffer,*count,cb_control,cookie);

	if(!status)
	{
		if(acquire_sem_etc(ad->notify,1,B_CAN_INTERRUPT|B_RELATIVE_TIMEOUT,100000000))
		{
			/* timeout or abort */
			usb->cancel_queued_transfers(ad->out);
			DPRINTF((ID "INTERRUPTED OR TIMEOUT\n"));
			acquire_sem(ad->notify);
			ad->actual=0;
		}
		DPRINTF((ID " data written(%ld)\n",ad->actual ));
//		memcpy((void *)buf,ad->buffer,ad->actual);
		*count = ad->actual;
	}
		release_sem(ad->lock);

	return ad->status;

//#if 0
//#endif
//		
//	if(channel->remain == 0) {
//		st = acquire_sem_etc(channel->num_available_packets, 1, B_RELATIVE_TIMEOUT, 4*1000*1000);
//		if(st) {
//			if(st == B_TIMED_OUT) {
//				dprintf("st = B_TIMED_OUT\n");
//				*count = 0;
//			}
//			return st;
//		}
//	
//		if(channel->current_rw == channel->next_to_queue) dprintf("e");
//		
//		channel->remain = channel->current_rw->buffer_size;
//		channel->buf_ptr = channel->current_rw->buffer;
//	}
//	
//	if(channel->remain >= *count) {
//		memcpy(channel->buf_ptr, buf, *count);
//		channel->remain -= *count;
//		channel->buf_ptr += *count;
//	} else {
//		memcpy(channel->buf_ptr, buf, channel->remain);
//		*count = channel->remain;
//		channel->remain = 0;
//		channel->buf_ptr = NULL;
//	}
//	
//	return B_OK;
}

static status_t
device_control (void *cookie, uint32 msg, void *arg1, size_t len)
{
	scannerdev* ad = (scannerdev*) cookie; 

	// if it's the first time through, open the fake fd used to avoid a 0
	// open count and so driver uniinitialization
	
	if(ad->fake_fd==-2) 
	{
		ad->fake_fd=open("/dev/scanner/usb/0",0);
	}

	if(msg==(B_DEVICE_OP_CODES_END+1))
	{
		// get vendor id
		const usb_device_descriptor *dd;
		
		dd= usb->get_device_descriptor(ad->dev);
		DPRINTF((ID "got vendor id %d\n",dd->vendor_id));
		memcpy(arg1,&(dd->vendor_id),2);
		
		return(B_OK);
	}
	
	return B_ERROR;
}


/* Driver Hooks ---------------------------------------------------------
**
** These functions provide the glue used by DevFS to load/unload
** the driver and also handle registering with the USB bus manager
** to receive device added and removed events
*/

static usb_notify_hooks notify_hooks = 
{
	&device_added,
	&device_removed
};

usb_support_descriptor supported_devices[] =
{
	/* Acer */
	{ 0, 0, 0, 0x04a5, 0x1a20 },	/* Prisa 310U */
	{ 0, 0, 0, 0x04a5, 0x1a2a },	/* Another 620U */
	{ 0, 0, 0, 0x04a5, 0x2060 },	/* Prisa Acerscan 620U & 640U (!)*/
	{ 0, 0, 0, 0x04a5, 0x2040 },	/* Prisa AcerScan 620U (!) */
	{ 0, 0, 0, 0x04a5, 0x2022 },	/* Vuego Scan Brisa 340U */
	{ 0, 0, 0, 0x04a5, 0x207e },	/* 640BU */
	{ 0, 0, 0, 0x04a5, 0x20b0 },	/* Benq 4300 */
	{ 0, 0, 0, 0x04a5, 0x20be },	/* Unknown - Oliver Schwartz */
	{ 0, 0, 0, 0x04a5, 0x20c0 }, 	/* 1240UT, 1240U */
	{ 0, 0, 0, 0x04a5, 0x20de },	/* S2W 3300U */
	{ 0, 0, 0, 0x04a5, 0x20fc }, 	/* Benq 5000 */
	{ 0, 0, 0, 0x04a5, 0x20fe },	/* Benq 5300 */
	/* Agfa */
	{ 0, 0, 0, 0x06bd, 0x0001 },	/* SnapScan 1212U */
	{ 0, 0, 0, 0x06bd, 0x0002 },	/* SnapScan 1236U */
	{ 0, 0, 0, 0x06bd, 0x208d }, 	/* Snapscan e40 */
	{ 0, 0, 0, 0x06bd, 0x208f }, 	/* SnapScan e50*/
	{ 0, 0, 0, 0x06bd, 0x2091 },	/* SnapScan e20 */
	{ 0, 0, 0, 0x06bd, 0x2093 }, 	/* SnapScan e10*/
	{ 0, 0, 0, 0x06bd, 0x2095 }, 	/* SnapScan e25 */
	{ 0, 0, 0, 0x06bd, 0x2097 }, 	/* SnapScan e26 */
	{ 0, 0, 0, 0x06bd, 0x20fd }, 	/* SnapScan e52*/
	{ 0, 0, 0, 0x06bd, 0x20ff }, 	/* SnapScan e42*/
	{ 0, 0, 0, 0x06bd, 0x2061 },	/* Another SnapScan 1212U (?)*/
	{ 0, 0, 0, 0x06bd, 0x0100 },	/* SnapScan Touch */
	/* Artec */
	{ 0, 0, 0, 0x05d8, 0x4001 },	/* Ultima 2000 */
	{ 0, 0, 0, 0x05d8, 0x4002 }, 	/* Ultima 2000 (GT6801 based) */
	{ 0, 0, 0, 0x05d8, 0x4003 }, 	/* E+ 48U */
	{ 0, 0, 0, 0x05d8, 0x4004 }, 	/* E+ Pro */
	/* Avision */
	{ 0, 0, 0, 0x0638, 0x0268 }, 	/* iVina 1200U */
	{ 0, 0, 0, 0x0638, 0x0a10 },	/* iVina FB1600 (=Umax Astra 4500) */
	{ 0, 0, 0, 0x0638, 0x0a20 }, 	/* iVina FB1800 (=Umax Astra 4700) */
	/* Benq: see Acer */
	/* Brother */
	{ 0, 0, 0, 0x04f9, 0x010f },	/* MFC 5100C */
	{ 0, 0, 0, 0x04f9, 0x0111 },	/* MFC 6800 */
	/* Canon */
	{ 0, 0, 0, 0x04a9, 0x2201 }, 	/* CanoScan FB320U */
	{ 0, 0, 0, 0x04a9, 0x2202 }, 	/* CanoScan FB620U */
	{ 0, 0, 0, 0x04a9, 0x2204 }, 	/* CanoScan FB630U/FB636U */
	{ 0, 0, 0, 0x04a9, 0x2205 }, 	/* CanoScan FB1210U */
	{ 0, 0, 0, 0x04a9, 0x2206 }, 	/* CanoScan N650U/N656U */
	{ 0, 0, 0, 0x04a9, 0x2207 }, 	/* CanoScan N1220U */
	{ 0, 0, 0, 0x04a9, 0x2208 }, 	/* CanoScan D660U */ 
	{ 0, 0, 0, 0x04a9, 0x220a },	/* CanoScan D2400UF */
	{ 0, 0, 0, 0x04a9, 0x220b }, 	/* CanoScan D646U */
	{ 0, 0, 0, 0x04a9, 0x220c },	/* CanoScan D1250U2 */
	{ 0, 0, 0, 0x04a9, 0x220d }, 	/* CanoScan N670U/N676U/LIDE 20 */
	{ 0, 0, 0, 0x04a9, 0x220e }, 	/* CanoScan N1240U/LIDE 30 */
	{ 0, 0, 0, 0x04a9, 0x220f },	/* CanoScan 8000F */
	{ 0, 0, 0, 0x04a9, 0x2213 },	/* LIDE 50 */
	{ 0, 0, 0, 0x04a9, 0x3042 }, 	/* FS4000US */
	/* Colorado -- See Primax/Colorado below */
	/* Compaq */
	{ 0, 0, 0, 0x049f, 0x001a },	/* S4 100 */
	{ 0, 0, 0, 0x049f, 0x0021 },	/* S200 */
	/* Epson -- See Seiko/Epson below */
	/* Fujitsu */
	{ 0, 0, 0, 0x04c5, 0x1041 }, 	/* fi-4220c USB/SCSI info:mza@mu-tec.de */
	{ 0, 0, 0, 0x04c5, 0x1042 }, 	/* fi-4120c USB/SCSI info:mza@mu-tec.de */
	{ 0, 0, 0, 0x04c5, 0x1029 }, 	/* fi-4010c USB AVision info:mza@mu-tec.de */
	/* Genius */
	{ 0, 0, 0, 0x0458, 0x2001 },	/* ColorPage-Vivid Pro */
	{ 0, 0, 0, 0x0458, 0x2007 },	/* ColorPage HR6 V2 */
	{ 0, 0, 0, 0x0458, 0x2008 }, 	/* ColorPage HR6 V2 */
	{ 0, 0, 0, 0x0458, 0x2009 }, 	/* ColorPage HR6A */
	{ 0, 0, 0, 0x0458, 0x2011 }, 	/* ColorPage Vivid3x */
	{ 0, 0, 0, 0x0458, 0x2013 }, 	/* ColorPage HR7 */
	{ 0, 0, 0, 0x0458, 0x2015 }, 	/* ColorPage HR7LE */
	{ 0, 0, 0, 0x0458, 0x2016 }, 	/* ColorPage HR6X */
	{ 0, 0, 0, 0x0458, 0x2018 },	/* ColorPage HR7X */
	/* Hewlett Packard */
	{ 0, 0, 0, 0x03f0, 0x0205 },	/* 3300C */
	{ 0, 0, 0, 0x03f0, 0x0101 },	/* 4100C */
	{ 0, 0, 0, 0x03f0, 0x0105 },	/* 4200C */
	{ 0, 0, 0, 0x03f0, 0x0102 },	/* PhotoSmart S20 */
	{ 0, 0, 0, 0x03f0, 0x0401 },	/* 5200C */
	{ 0, 0, 0, 0x03f0, 0x0701 },	/* 5300C */
	{ 0, 0, 0, 0x03f0, 0x0201 },	/* 6200C */
	{ 0, 0, 0, 0x03f0, 0x0601 },	/* 6300C */
	{ 0, 0, 0, 0x03f0, 0x0305 }, 	/* ScanJet 4300C */
	{ 0, 0, 0, 0x03f0, 0x0405 }, 	/* ScanJet 3400C */
	{ 0, 0, 0, 0x03f0, 0x0505 }, 	/* ScanJet 2100C */
	{ 0, 0, 0, 0x03f0, 0x0605 },	/* ScanJet 2200C */
	{ 0, 0, 0, 0x03f0, 0x0705 }, 	/* ScanJet 4400C */
	{ 0, 0, 0, 0x03f0, 0x0901 }, 	/* ScanJet 2300C */
	{ 0, 0, 0, 0x03F0, 0x1005 },	/* ScanJet 5400C */
	{ 0, 0, 0, 0x03F0, 0x1105 },	/* ScanJet 5470C */
	{ 0, 0, 0, 0x03f0, 0x1205 }, 	/* ScanJet 5550C */
	{ 0, 0, 0, 0x03f0, 0x1305 },	/* Scanjet 4570c */
	{ 0, 0, 0, 0x03f0, 0x2005 },	/* ScanJet 3570c */
	{ 0, 0, 0, 0x03f0, 0x2205 },	/* ScanJet 3500c */
	/* Lexmark */
	{ 0, 0, 0, 0x043d, 0x002d }, 	/* X70/X73 */
	{ 0, 0, 0, 0x043d, 0x003d }, 	/* X83 */
	/* LG Electronics */
	{ 0, 0, 0, 0x0461, 0x0364 },	/* Scanworks 600U (repackaged Primax?) */
	/* Medion */
	{ 0, 0, 0, 0x0461, 0x0377 },	/* MD 5345 - repackaged Primax? */
	/* Memorex */
	{ 0, 0, 0, 0x0461, 0x0346 }, 	/* 6136u - repackaged Primax ? */
	/* iVina */
	{ 0, 0, 0, 0x0638, 0x0268 },     /* 1200U */
	/* Microtek */
	{ 0, 0, 0, 0x05da, 0x0099 },	/* ScanMaker X6 - X6U */
	{ 0, 0, 0, 0x05da, 0x0094 },	/* Phantom 336CX - C3 */
	{ 0, 0, 0, 0x05da, 0x00a0 },	/* Phantom 336CX - C3 #2 */
	{ 0, 0, 0, 0x05da, 0x009a },	/* Phantom C6 */
	{ 0, 0, 0, 0x05da, 0x00a3 },	/* ScanMaker V6USL */
	{ 0, 0, 0, 0x05da, 0x80a3 },	/* ScanMaker V6USL #2 */
	{ 0, 0, 0, 0x05da, 0x80ac },	/* ScanMaker V6UL - SpicyU */
	{ 0, 0, 0, 0x05da, 0x20c9 }, 	/* ScanMaker 6700 */
	{ 0, 0, 0, 0x05da, 0x30ce },	/* ScanMaker 3800 */
	{ 0, 0, 0, 0x05da, 0x30cf },	/* ScanMaker 4800 */
	{ 0, 0, 0, 0x04a7, 0x0224 },	/* Scanport 3000 (actually Visioneer?)*/
	/* Minolta */
	{ 0, 0, 0, 0x0686, 0x400d }, 	/* Scan Dual III */
	/* Mustek */
	{ 0, 0, 0, 0x0400, 0x1001 }, 	/* BearPaw 2400 (National Semiconductor LM9832) */
	{ 0, 0, 0, 0x055f, 0x0001 },	/* 1200 CU */
	{ 0, 0, 0, 0x0400, 0x1000 },	/* BearPaw 1200 */
	{ 0, 0, 0, 0x055f, 0x0002 },	/* 600 CU */
	{ 0, 0, 0, 0x055f, 0x0003 },	/* 1200 USB */
	{ 0, 0, 0, 0x055f, 0x0006 },	/* 1200 UB */
	{ 0, 0, 0, 0x055f, 0x0007 },	/* ScanExpress 1200 USB Plus */
	{ 0, 0, 0, 0x055f, 0x0008 }, 	/* ScanExpress 1200 CU Plus */
	{ 0, 0, 0, 0x055f, 0x0010 }, 	/* BearPaw 1200F */
	{ 0, 0, 0, 0x055f, 0x0210 },	/* ScanExpress A3 USB */
	{ 0, 0, 0, 0x055f, 0x0218 }, 	/* BearPaw 2400 TA */
	{ 0, 0, 0, 0x055f, 0x0219 }, 	/* BearPaw 2400 TA Plus */
	{ 0, 0, 0, 0x055f, 0x021c }, 	/* BearPaw 1200 CU Plus */
	{ 0, 0, 0, 0x055f, 0x021d }, 	/* Bearpaw 2400 CU Plus */
	{ 0, 0, 0, 0x055f, 0x021e }, 	/* BearPaw 1200 TA/CS */
	{ 0, 0, 0, 0x055f, 0x0400 }, 	/* BearPaw 2400 TA PRO */
	{ 0, 0, 0, 0x055f, 0x0401 },	/* P 3600 A3 Pro */
	{ 0, 0, 0, 0x055f, 0x0873 }, 	/* ScanExpress 600 USB */
	{ 0, 0, 0, 0x055f, 0x1000 }, 	/* BearPaw 4800 TA PRO */
	/* Nikon */
	{ 0, 0, 0, 0x04b0, 0x4000 }, 	/* Coolscan LS 40 ED */
	/* Pacific Image Electronics */
	{ 0, 0, 0, 0x05e3, 0x0120 },	/* PrimeFilm 1800u */
	/* Plustek */
	{ 0, 0, 0, 0x07b3, 0x0001 },	/* 1212U */
	{ 0, 0, 0, 0x07b3, 0x0005 }, 	/* Unknown */
	{ 0, 0, 0, 0x07b3, 0x0007 }, 	/* Unknown */
	{ 0, 0, 0, 0x07b3, 0x000F }, 	/* Unknown */
	{ 0, 0, 0, 0x07b3, 0x0010 }, 	/* OpticPro U12 */
	{ 0, 0, 0, 0x07b3, 0x0011 }, 	/* OpticPro U24 */
	{ 0, 0, 0, 0x07b3, 0x0012 }, 	/* Unknown */
	{ 0, 0, 0, 0x07b3, 0x0013 }, 	/* UT12 */
	{ 0, 0, 0, 0x07b3, 0x0014 }, 	/* Unknown */
	{ 0, 0, 0, 0x07b3, 0x0015 }, 	/* OpticPro U24 */
	{ 0, 0, 0, 0x07b3, 0x0016 }, 	/* Unknown */
	{ 0, 0, 0, 0x07b3, 0x0017 }, 	/* OpticPro UT12/UT16/UT24 */
	{ 0, 0, 0, 0x07b3, 0x0400 }, 	/* OpticPro 1248U */
	{ 0, 0, 0, 0x07b3, 0x0401 }, 	/* OpticPro 1248U (another one) */
	{ 0, 0, 0, 0x07b3, 0x0403 },	/* U16B */
	/* Primax/Colorado */
	{ 0, 0, 0, 0x0461, 0x0300 },	/* G2-300 #1 */
	{ 0, 0, 0, 0x0461, 0x0380 },	/* G2-600 #1 */
	{ 0, 0, 0, 0x0461, 0x0301 },	/* G2E-300 #1 */
	{ 0, 0, 0, 0x0461, 0x0381 },	/* ReadyScan 636i */
	{ 0, 0, 0, 0x0461, 0x0302 },	/* G2-300 #2 */
	{ 0, 0, 0, 0x0461, 0x0382 },	/* G2-600 #2 */
	{ 0, 0, 0, 0x0461, 0x0303 },	/* G2E-300 #2 */
	{ 0, 0, 0, 0x0461, 0x0383 },	/* G2E-600 */
	{ 0, 0, 0, 0x0461, 0x0340 },	/* Colorado USB 9600 */
	{ 0, 0, 0, 0x0461, 0x0360 },	/* Colorado USB 19200 */
	{ 0, 0, 0, 0x0461, 0x0341 },	/* Colorado 600u */
	{ 0, 0, 0, 0x0461, 0x0361 },	/* Colorado 1200u */
	{ 0, 0, 0, 0x0461, 0x0347 },	/* Primascan Colorado 2600u */
	/* Prolink */
	{ 0, 0, 0, 0x06dc, 0x0014 }, 	/* Winscan Pro 2448U */
	/* Relisis */
	// { USB_DEVICE(0x0475, 0x0103) },	/* Episode - undetected endpoint */
	{ 0, 0, 0, 0x0475, 0x0210 }, 	/* Scorpio Ultra 3 */
	/* Seiko/Epson Corp. */
	{ 0, 0, 0, 0x04b8, 0x0101 },	/* Perfection 636U and 636Photo */
	{ 0, 0, 0, 0x04b8, 0x0102 }, 	/* GT-2200 */
	{ 0, 0, 0, 0x04b8, 0x0103 },	/* Perfection 610 */
	{ 0, 0, 0, 0x04b8, 0x010c }, 	/* Perfection 640U */
	{ 0, 0, 0, 0x04b8, 0x0104 },	/* Perfection 1200U and 1200Photo*/
	{ 0, 0, 0, 0x04b8, 0x0105 }, 	/* StylusScan 2000 */
	{ 0, 0, 0, 0x04b8, 0x0106 },	/* Stylus Scan 2500 */
	{ 0, 0, 0, 0x04b8, 0x0107 },	/* Expression 1600 */
	{ 0, 0, 0, 0x04b8, 0x0109 }, 	/* Expression 1640XL */
	{ 0, 0, 0, 0x04b8, 0x010a }, 	/* Perfection 1640SU and 1640SU Photo */
	{ 0, 0, 0, 0x04b8, 0x010b }, 	/* Perfection 1240U */
	{ 0, 0, 0, 0x04b8, 0x010e }, 	/* Expression 1680 */
	{ 0, 0, 0, 0x04b8, 0x010f }, 	/* Perfection 1250U */
	{ 0, 0, 0, 0x04b8, 0x0110 }, 	/* Perfection 1650 */
	{ 0, 0, 0, 0x04b8, 0x0112 }, 	/* Perfection 2450 - GT-9700 for the Japanese mkt */
	{ 0, 0, 0, 0x04b8, 0x0114 }, 	/* Perfection 660 */
	{ 0, 0, 0, 0x04b8, 0x011b }, 	/* Perfection 2400 Photo */
	{ 0, 0, 0, 0x04b8, 0x011c }, 	/* Perfection 3200 */
	{ 0, 0, 0, 0x04b8, 0x011d }, 	/* Perfection 1260 */
	{ 0, 0, 0, 0x04b8, 0x011e }, 	/* Perfection 1660 Photo */
	{ 0, 0, 0, 0x04b8, 0x0802 }, 	/* Stylus CX3200 */
	{ 0, 0, 0, 0x04b8, 0x0801 }, 	/* Stylus CX5200 */
	{ 0, 0, 0, 0x04b8, 0x0808 }, 	/* Stylus CX5400 */
	{ 0, 0, 0, 0x04b8, 0x0805 }, 	/* Stylus CX6400 */
	/* SYSCAN */
	{ 0, 0, 0, 0x0a82, 0x4600 }, 	/* TravelScan 460/464 */
	/* Trust */
	{ 0, 0, 0, 0x05cb, 0x1483 }, 	/* CombiScan 19200 */
	{ 0, 0, 0, 0x05d8, 0x4006 }, 	/* Easy Webscan 19200 (repackaged Artec?) */
	/* Umax */
	{ 0, 0, 0, 0x05d8, 0x4009 },	/* Astraslim (actually Artec?) */
	{ 0, 0, 0, 0x1606, 0x0060 }, 	/* Astra 3400U/3450U */
	{ 0, 0, 0, 0x1606, 0x0130 },	/* Astra 2100U */
	{ 0, 0, 0, 0x1606, 0x0160 }, 	/* Astra 5400U */
	{ 0, 0, 0, 0x1606, 0x0010 },	/* Astra 1220U */
	{ 0, 0, 0, 0x1606, 0x0030 },	/* Astra 2000U */
	{ 0, 0, 0, 0x1606, 0x0230 },	/* Astra 2200U */
	/* Visioneer */
	{ 0, 0, 0, 0x04a7, 0x0221 },	/* OneTouch 5300 USB */
	{ 0, 0, 0, 0x04a7, 0x0211 },	/* OneTouch 7600 USB */
	{ 0, 0, 0, 0x04a7, 0x0224 },	/* OneTouch 4800 USB */
	{ 0, 0, 0, 0x04a7, 0x0226 },	/* OneTouch 5300 USB */
	{ 0, 0, 0, 0x04a7, 0x0231 },	/* 6100 USB */
	{ 0, 0, 0, 0x04a7, 0x0311 },	/* 6200 EPP/USB */
	{ 0, 0, 0, 0x04a7, 0x0321 },	/* OneTouch 8100 EPP/USB */
	{ 0, 0, 0, 0x04a7, 0x0331 },	/* OneTouch 8600 EPP/USB */
	{ 0, 0, 0, 0x0461, 0x0345 }, 	/* 6200 (actually Primax?) */
	{ 0, 0, 0, 0x0461, 0x0371 }, 	/* Onetouch 8920 USB (actually Primax?) */
};

_EXPORT status_t 
init_hardware(void)
{
	return B_OK;
}

_EXPORT status_t
init_driver(void)
{
	int i;
	DPRINTF((ID "init_driver(), built %s %s\n", __DATE__, __TIME__));
	
#if DEBUG_DRIVER
	if(load_driver_symbols(drivername) == B_OK) {
		DPRINTF((ID "loaded symbols\n"));
	} else {
		DPRINTF((ID "no symbols for you!\n"));
	}
#endif
			
	if(get_module(usb_name,(module_info**) &usb) != B_OK){
		dprintf(ID "cannot get module \"%s\"\n",usb_name);
		return B_ERROR;
	} 
	
	if((dev_list_lock = create_sem(1,"dev_list_lock")) < 0){
		put_module(usb_name);
		return dev_list_lock;
	}
	
	usb->register_driver(drivername, supported_devices, 194, NULL);
	usb->install_notify(drivername, &notify_hooks);
	
	return B_OK;
}

_EXPORT void
uninit_driver(void)
{
	int i;
	
	DPRINTF((ID "uninit_driver()\n"));
	
	usb->uninstall_notify(drivername);
	
	delete_sem(dev_list_lock);

	put_module(usb_name);

	if(device_names){
		for(i=0;device_names[i];i++) free(device_names[i]);
		free(device_names);
	}
	
}

_EXPORT const char**
publish_devices()
{
	scannerdev *ad;
	int i;
	
	DPRINTF((ID "publish_devices()\n"));
	
	if(device_names){
		for(i=0;device_names[i];i++) free((char *) device_names[i]);
		free(device_names);
	}
	
	acquire_sem(dev_list_lock);	
	device_names = (char **) malloc(sizeof(char*) * (device_count + 1));	
	if(device_names){
		for(i = 0, ad = device_list; ad; ad = ad->next){
//			if(ad->out_channel.active)
			{
				if(device_names[i] = (char *) malloc(strlen(basename) + 4)){
//					sprintf(device_names[i],"%s/%d",basename,ad->number);
					sprintf(device_names[i],"%s/0",basename);
					DPRINTF((ID "publishing: \"/dev/%s\"\n",device_names[i]));
					i++;
				}
			}
		}
		device_names[i] = NULL;
	}
	release_sem(dev_list_lock);
	
	return (const char **) device_names;
}

static device_hooks DeviceHooks = {
	device_open, 			
	device_close, 			
	device_free,			
	device_control, 		
	device_read,			
	device_write			 
};

_EXPORT device_hooks*
find_device(const char* name)
{
	return &DeviceHooks;
}
