// *** LICENSE ***
// This file is Copyright (c) 2001 Marcus Overhagen
//
// It is being made publicly available and free to use
// in freeware, shareware or commercial products.
// 
// If you use this software or parts of it in your 
// own product, you are required to state in your 
// documentation and about box or copyright notice 
// that portions of your sofware were developed
// by Marcus Overhagen.
//
// This software is provided 'as-is', without any 
// express or implied warranty. In no event will 
// the author be held liable for any damages (direct,
// indirect or otherwise) arising from the use of
// this software. In the case of states in which 
// such disclaimers do not apply, the author hereby
// limits his liability in such matters to USD $1. 

#include <Application.h>
#include <Window.h>
#include <View.h>
#include <Cursor.h>
#include "Debug.h"
#include "SplitterView.h"

	static const uint8 split_cursor[] = { 16,1,7,7,
		0 <<7| 0 <<6| 0 <<5| 0 <<4| 0 <<3| 0 <<2| 0 <<1| 0, 0 <<7| 0 <<6| 0 <<5| 0 <<4| 0 <<3| 0 <<2| 0 <<1| 0,
		0 <<7| 0 <<6| 0 <<5| 0 <<4| 0 <<3| 0 <<2| 0 <<1| 1, 1 <<7| 0 <<6| 0 <<5| 0 <<4| 0 <<3| 0 <<2| 0 <<1| 0, 
		0 <<7| 0 <<6| 0 <<5| 0 <<4| 0 <<3| 0 <<2| 0 <<1| 1, 1 <<7| 0 <<6| 0 <<5| 0 <<4| 0 <<3| 0 <<2| 0 <<1| 0,
		0 <<7| 0 <<6| 0 <<5| 0 <<4| 0 <<3| 0 <<2| 0 <<1| 1, 1 <<7| 0 <<6| 0 <<5| 0 <<4| 0 <<3| 0 <<2| 0 <<1| 0,
		0 <<7| 0 <<6| 0 <<5| 0 <<4| 0 <<3| 0 <<2| 0 <<1| 1, 1 <<7| 0 <<6| 0 <<5| 0 <<4| 0 <<3| 0 <<2| 0 <<1| 0, 
		0 <<7| 0 <<6| 0 <<5| 1 <<4| 0 <<3| 0 <<2| 0 <<1| 1, 1 <<7| 0 <<6| 0 <<5| 0 <<4| 1 <<3| 0 <<2| 0 <<1| 0, 
		0 <<7| 0 <<6| 1 <<5| 1 <<4| 0 <<3| 0 <<2| 0 <<1| 1, 1 <<7| 0 <<6| 0 <<5| 0 <<4| 1 <<3| 1 <<2| 0 <<1| 0, 
		0 <<7| 1 <<6| 1 <<5| 1 <<4| 1 <<3| 1 <<2| 1 <<1| 1, 1 <<7| 1 <<6| 1 <<5| 1 <<4| 1 <<3| 1 <<2| 1 <<1| 0, 
		0 <<7| 0 <<6| 1 <<5| 1 <<4| 0 <<3| 0 <<2| 0 <<1| 1, 1 <<7| 0 <<6| 0 <<5| 0 <<4| 1 <<3| 1 <<2| 0 <<1| 0, 
		0 <<7| 0 <<6| 0 <<5| 1 <<4| 0 <<3| 0 <<2| 0 <<1| 1, 1 <<7| 0 <<6| 0 <<5| 0 <<4| 1 <<3| 0 <<2| 0 <<1| 0, 
		0 <<7| 0 <<6| 0 <<5| 0 <<4| 0 <<3| 0 <<2| 0 <<1| 1, 1 <<7| 0 <<6| 0 <<5| 0 <<4| 0 <<3| 0 <<2| 0 <<1| 0, 
		0 <<7| 0 <<6| 0 <<5| 0 <<4| 0 <<3| 0 <<2| 0 <<1| 1, 1 <<7| 0 <<6| 0 <<5| 0 <<4| 0 <<3| 0 <<2| 0 <<1| 0, 
		0 <<7| 0 <<6| 0 <<5| 0 <<4| 0 <<3| 0 <<2| 0 <<1| 1, 1 <<7| 0 <<6| 0 <<5| 0 <<4| 0 <<3| 0 <<2| 0 <<1| 0, 
		0 <<7| 0 <<6| 0 <<5| 0 <<4| 0 <<3| 0 <<2| 0 <<1| 1, 1 <<7| 0 <<6| 0 <<5| 0 <<4| 0 <<3| 0 <<2| 0 <<1| 0, 
		0 <<7| 0 <<6| 0 <<5| 0 <<4| 0 <<3| 0 <<2| 0 <<1| 0, 0 <<7| 0 <<6| 0 <<5| 0 <<4| 0 <<3| 0 <<2| 0 <<1| 0,
		0 <<7| 0 <<6| 0 <<5| 0 <<4| 0 <<3| 0 <<2| 0 <<1| 0, 0 <<7| 0 <<6| 0 <<5| 0 <<4| 0 <<3| 0 <<2| 0 <<1| 0, 
		0 <<7| 0 <<6| 0 <<5| 0 <<4| 0 <<3| 0 <<2| 1 <<1| 1, 1 <<7| 1 <<6| 0 <<5| 0 <<4| 0 <<3| 0 <<2| 0 <<1| 0, 
		0 <<7| 0 <<6| 0 <<5| 0 <<4| 0 <<3| 0 <<2| 1 <<1| 1, 1 <<7| 1 <<6| 0 <<5| 0 <<4| 0 <<3| 0 <<2| 0 <<1| 0, 
		0 <<7| 0 <<6| 0 <<5| 0 <<4| 0 <<3| 0 <<2| 1 <<1| 1, 1 <<7| 1 <<6| 0 <<5| 0 <<4| 0 <<3| 0 <<2| 0 <<1| 0, 
		0 <<7| 0 <<6| 0 <<5| 0 <<4| 0 <<3| 0 <<2| 1 <<1| 1, 1 <<7| 1 <<6| 0 <<5| 0 <<4| 0 <<3| 0 <<2| 0 <<1| 0, 
		0 <<7| 0 <<6| 0 <<5| 1 <<4| 0 <<3| 0 <<2| 1 <<1| 1, 1 <<7| 1 <<6| 0 <<5| 0 <<4| 1 <<3| 0 <<2| 0 <<1| 0, 
		0 <<7| 0 <<6| 1 <<5| 1 <<4| 1 <<3| 0 <<2| 1 <<1| 1, 1 <<7| 1 <<6| 0 <<5| 1 <<4| 1 <<3| 1 <<2| 0 <<1| 0, 
		0 <<7| 1 <<6| 1 <<5| 1 <<4| 1 <<3| 1 <<2| 1 <<1| 1, 1 <<7| 1 <<6| 1 <<5| 1 <<4| 1 <<3| 1 <<2| 1 <<1| 0, 
		1 <<7| 1 <<6| 1 <<5| 1 <<4| 1 <<3| 1 <<2| 1 <<1| 1, 1 <<7| 1 <<6| 1 <<5| 1 <<4| 1 <<3| 1 <<2| 1 <<1| 1, 
		0 <<7| 1 <<6| 1 <<5| 1 <<4| 1 <<3| 1 <<2| 1 <<1| 1, 1 <<7| 1 <<6| 1 <<5| 1 <<4| 1 <<3| 1 <<2| 1 <<1| 0, 
		0 <<7| 0 <<6| 1 <<5| 1 <<4| 1 <<3| 0 <<2| 1 <<1| 1, 1 <<7| 1 <<6| 0 <<5| 1 <<4| 1 <<3| 1 <<2| 0 <<1| 0, 
		0 <<7| 0 <<6| 0 <<5| 1 <<4| 0 <<3| 0 <<2| 1 <<1| 1, 1 <<7| 1 <<6| 0 <<5| 0 <<4| 1 <<3| 0 <<2| 0 <<1| 0, 
		0 <<7| 0 <<6| 0 <<5| 0 <<4| 0 <<3| 0 <<2| 1 <<1| 1, 1 <<7| 1 <<6| 0 <<5| 0 <<4| 0 <<3| 0 <<2| 0 <<1| 0, 
		0 <<7| 0 <<6| 0 <<5| 0 <<4| 0 <<3| 0 <<2| 1 <<1| 1, 1 <<7| 1 <<6| 0 <<5| 0 <<4| 0 <<3| 0 <<2| 0 <<1| 0, 
		0 <<7| 0 <<6| 0 <<5| 0 <<4| 0 <<3| 0 <<2| 1 <<1| 1, 1 <<7| 1 <<6| 0 <<5| 0 <<4| 0 <<3| 0 <<2| 0 <<1| 0, 
		0 <<7| 0 <<6| 0 <<5| 0 <<4| 0 <<3| 0 <<2| 1 <<1| 1, 1 <<7| 1 <<6| 0 <<5| 0 <<4| 0 <<3| 0 <<2| 0 <<1| 0, 
		0 <<7| 0 <<6| 0 <<5| 0 <<4| 0 <<3| 0 <<2| 0 <<1| 0, 0 <<7| 0 <<6| 0 <<5| 0 <<4| 0 <<3| 0 <<2| 0 <<1| 0
	};


CSplitterView::CSplitterView(BRect frame, const char *name, BView *left, BView *right, uint32 resizingMode, uint32 flags) 
	: BView(frame,name,resizingMode,flags/* | B_FULL_UPDATE_ON_RESIZE */ | B_FRAME_EVENTS | B_WILL_DRAW)
{
	fPercent = 50;
	bIsTracking = false;
	bHasTrackingCursor = false;
	viewLeft = left;
	viewRight = right;	

	SetViewColor(216,216,216);
	//SetViewColor(0,0,0);

	cursor = new BCursor(split_cursor);

	viewLeft->SetResizingMode(B_FOLLOW_NONE);
	viewRight->SetResizingMode(B_FOLLOW_NONE);
	viewLeft->MoveTo(0,0);
	AddChild(viewLeft);
	AddChild(viewRight);

	FrameResized(Bounds().Width(), Bounds().Height());
}

CSplitterView::~CSplitterView()
{
	RemoveChild(viewLeft);
	RemoveChild(viewRight);
	//delete viewLeft;
	//delete viewRight;
	delete cursor;
}

void CSplitterView::FrameResized(float width, float height)
{
	float widthL;
	float widthR;
	BWindow *win;

	win = Window();
	widthL = (float)(int)((width * fPercent) / 100); // Nachkommastellen abschneiden
	widthR = width - widthL;
	
	if (win) {
		win->DisableUpdates();
		win->BeginViewTransaction();
	}
	if (fPercent > 0) {
		viewLeft->MoveTo(0,0);
		viewLeft->ResizeTo(widthL - (X_BAR_SIZE / 2),height);
		viewRight->MoveTo(widthL + 1 + (X_BAR_SIZE / 2),0);
		viewRight->ResizeTo(widthR - (X_BAR_SIZE / 2) - 1,height);
	} else {
		viewLeft->MoveTo(-10000,-10000);
		viewLeft->ResizeTo(200,200);
		viewRight->MoveTo(0,0);
		viewRight->ResizeTo(width,height);
	}
	if (win) {
		win->EndViewTransaction();
		win->EnableUpdates();
		Sync();
	}
}

void CSplitterView::MouseDown(BPoint p)
{
	BRect rect(Bounds());
	float widthL;
	BWindow *win;

	win = Window();
	if (win == 0 || false == win->IsActive())
		return;

	widthL = (float)(int)((rect.Width() * fPercent) / 100); // Nachkommastellen abschneiden
	
	if (p.y >= 0 && p.y <= rect.Height() && p.x >= widthL - (X_BAR_SIZE / 2) && p.x <= widthL + 1 + (X_BAR_SIZE / 2)) {
		bIsTracking = true;
		SetMouseEventMask(B_POINTER_EVENTS,B_NO_POINTER_HISTORY | B_SUSPEND_VIEW_FOCUS | B_LOCK_WINDOW_FOCUS);
		DoMouse(p.x);
	}
}

void CSplitterView::AllAttached()
{
	//SetViewCursor(cursor, true);
}

void CSplitterView::MouseMoved(BPoint p, uint32 transit, const BMessage *message)
{
	BRect rect(Bounds());
	BWindow *win = Window();
	float widthL;

	//SetViewCursor(cursor, true);
		
	if (win && win->IsActive()) {
		widthL = (float)(int)((rect.Width() * fPercent) / 100); // Nachkommastellen abschneiden
		if (p.y >= 0 && p.y <= rect.Height() && p.x >= widthL - (X_BAR_SIZE / 2) && p.x <= widthL + 1 + (X_BAR_SIZE / 2)) {
			if (!bHasTrackingCursor && message == 0) {
				bHasTrackingCursor = true;
				be_app->SetCursor(cursor);
				SetEventMask(B_POINTER_EVENTS);
			}
		} else {
			if (bHasTrackingCursor && !bIsTracking) {
				be_app->SetCursor(B_HAND_CURSOR);
				bHasTrackingCursor = false;
				SetEventMask(EventMask() & ~B_POINTER_EVENTS);
			}
		}
		if (bIsTracking)
			DoMouse(p.x);
	}
}

void CSplitterView::MouseUp(BPoint p)
{
	bool bWasTracking;

	bWasTracking = bIsTracking;
	bIsTracking = false;

	if (bHasTrackingCursor && bWasTracking)
		MouseMoved(p, 0, 0);
}

void CSplitterView::DoMouse(float x)
{
	BRect rect(Bounds());

	fPercent = (x * 100) / rect.Width();
	if (fPercent < 10)
		fPercent = 10;
	if (fPercent > 90)
		fPercent = 90;
	FrameResized(rect.Width(), rect.Height());
}

void CSplitterView::SetPercent(float percent)
{
	fPercent = percent;
	
	if (fPercent > 0) {
		if (viewLeft->IsHidden())
			viewLeft->Show();
	} else {
		if (false == viewLeft->IsHidden())
			viewLeft->Hide();
		if (Window() != 0)
			viewRight->MakeFocus();
	}
	FrameResized(Bounds().Width(), Bounds().Height());
}

float CSplitterView::Percent()
{
	return fPercent;
}

void CSplitterView::MakeFocus(bool focused)
{
	if (focused) {
		if (fPercent > 0)
			viewLeft->MakeFocus();
		else 
			viewRight->MakeFocus();
	} else {
		if (viewLeft->IsFocus())
			viewLeft->MakeFocus(false);
		if (viewRight->IsFocus())
			viewRight->MakeFocus(false);
	}
}
